/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.framework;

import android.os.*;
import android.taobao.atlas.bundleInfo.AtlasBundleInfoManager;
import android.taobao.atlas.bundleInfo.BundleListing;
import android.taobao.atlas.runtime.LowDiskException;
import android.taobao.atlas.runtime.RuntimeVariables;
import android.taobao.atlas.util.ApkUtils;
import android.taobao.atlas.util.FileUtils;
import android.taobao.atlas.util.log.impl.AtlasMonitor;
import android.taobao.atlas.versionInfo.BaselineInfoManager;
import android.util.Log;
import android.util.Pair;
import org.osgi.framework.Bundle;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.zip.ZipEntry;

/**
 * Created by guanjie on 16/2/22.
 */
public class BundleInstaller implements Callable{

    private static Handler sBundleHandler;
    private static MessageQueue.IdleHandler sIdleHandler;
    private static List<Pair<String,InstallListener>>    sIdleInstallBundles = new ArrayList<Pair<String,InstallListener>>();
    private static List<Pair<String,InstallListener>>    sDelayInstallBundles = new ArrayList<Pair<String,InstallListener>>();
    private InstallListener mListener;
    private String mLocation[];
    private InputStream[] mBundleSourceInputStream;
    private File[]   mBundleSourceFile;
    private boolean mTransitive;

    private InputStream mTmpBundleSourceInputStream;
    private File        mTmpBundleSourceFile;

    static{
        HandlerThread handlerThread = new HandlerThread("bundle_installer");
        handlerThread.start();
        sBundleHandler = new Handler(handlerThread.getLooper());
    }

    public synchronized static void createIdleInstallerIfNeed(){
        if(sIdleHandler==null){
            sIdleHandler = new MessageQueue.IdleHandler() {
                @Override
                public boolean queueIdle() {
                    if (sIdleInstallBundles.size() == 0 && sDelayInstallBundles.size() == 0) {
                        sIdleHandler = null;
                        return false;
                    }
                    Pair<String,InstallListener> bundleStruct = null;
                    if (sDelayInstallBundles.size() > 0) {
                        bundleStruct = sDelayInstallBundles.remove(0);
                    } else if (sIdleInstallBundles.size() > 0) {
                        bundleStruct = sIdleInstallBundles.remove(0);
                    }
                    if(bundleStruct!=null) {
                        String bundleName = bundleStruct.first;
                        InstallListener listener = bundleStruct.second;
                        BundleImpl impl = (BundleImpl) Atlas.getInstance().getBundle(bundleName);
                        if (impl == null || !impl.checkValidate()) {
                            Log.d("BundleInstaller", "idle install bundle : " + bundleName);
                            BundleInstallerFetcher.obtainInstaller().installTransitivelySync(new String[]{bundleName});
                            if (listener != null) {
                                listener.onFinished();
                            }
                            return true;
                        }
                    }
                    return true;
                }
            };
            sBundleHandler.post(new Runnable() {
                @Override
                public void run() {
                    Looper.myQueue().addIdleHandler(sIdleHandler);
                }
            });
        }
    }

    public static void startIdleInstall(final String location,InstallListener listener){
        createIdleInstallerIfNeed();
        if(location!=null){
            BundleImpl impl = (BundleImpl)Atlas.getInstance().getBundle(location);
            if(impl==null || !impl.checkValidate()){
                sIdleInstallBundles.add(new Pair<String, InstallListener>(location,listener));
            }
        }

    }

    public static void startDelayInstall(final String location,InstallListener listener)
    {
        createIdleInstallerIfNeed();
        if(location!=null){
            BundleImpl impl = (BundleImpl)Atlas.getInstance().getBundle(location);
            if(impl==null || !impl.checkValidate()){
                sDelayInstallBundles.add(new Pair<String, InstallListener>(location,listener));
            }
        }
    }

    BundleInstaller(){
    }

    void release(){
        mLocation = null;
        mBundleSourceInputStream = null;
        mBundleSourceFile = null;
        mListener = null;
        mTmpBundleSourceFile = null;
        mTmpBundleSourceInputStream = null;
        mTransitive = false;
    }

    /////////////////////////////////////安装内部bundle////////////////////////////////////
    public void installTransitivelyAsync(String[] location, InstallListener listener){
        if(location!=null){
            release();
            mTransitive = true;
            mLocation = location;
            mListener = listener;
            try {
                installBundleInternal(false);
            }catch(Throwable e){
                e.printStackTrace();
                if(mListener!=null){
                    mListener.onFinished();
                }
                BundleInstallerFetcher.recycle(this);
            }
        }
    }

//    /**
//     * 最高优先级的bundle安装
//     * @param location
//     */
//    public void installTransitivelyOnUiThreadSync(String[] location){
//        if (location!=null){
//            mTransitive = true;
//            mLocation = location;
//            try {
//                call();
//            } catch (Exception e) {
//                e.printStackTrace();
//            }
//            BundleInstallerFetcher.recycle(this);
//        }
//    }

    public void installTransitivelySync(String[] location){
        if(location!=null){
            release();
            mTransitive = true;
            mLocation = location;
            try {
                installBundleInternal(true);
            }catch(Throwable e){
                e.printStackTrace();
                BundleInstallerFetcher.recycle(this);
            }
        }
    }

    /////////////////////////////////////安装内部bundle////////////////////////////////////
    /************************************************************************************/

    /************************************************************************************/
    /////////////////////////////////////安装外部bundle////////////////////////////////////
    public void installAsync(final String[] location, final InputStream[] input, InstallListener listener){
        if(Thread.currentThread().getId() != Looper.getMainLooper().getThread().getId()){
            throw new RuntimeException("method executeOnUIThread is only prepared for UI Thread");
        }
        release();
        mListener = listener;
        mBundleSourceInputStream = input;
        mLocation = location;
        try {
            checkSrc(location,input);
            installBundleInternal(false);
        }catch(Throwable e){
            e.printStackTrace();
            if(mListener!=null){
                mListener.onFinished();
            }
            BundleInstallerFetcher.recycle(this);
        }
    }

    public void installAsync(final String[] location, final File[] input, InstallListener listener){
        if(Thread.currentThread().getId() != Looper.getMainLooper().getThread().getId()){
            throw new RuntimeException("method executeOnUIThread is only prepared for UI Thread");
        }
        release();
        mListener = listener;
        mBundleSourceFile = input;
        mLocation = location;
        try {
            checkSrc(location,input);
            installBundleInternal(false);
        }catch(Throwable e){
            e.printStackTrace();
            BundleInstallerFetcher.recycle(this);
        }
    }

    public void installSync(final String[] location, final InputStream[] input){
        try {
            release();
            mBundleSourceInputStream = input;
            mLocation = location;
            checkSrc(location, input);
            installBundleInternal(true);
        }catch(Throwable e){
            e.printStackTrace();
            BundleInstallerFetcher.recycle(this);
        }
    }

    public void installSync(final String[] location, final File[] input){
        try {
            release();
            mBundleSourceFile = input;
            mLocation = location;
            checkSrc(location,input);
            installBundleInternal(true);
        }catch(Throwable e){
            e.printStackTrace();
            BundleInstallerFetcher.recycle(this);
        }
    }

//    /**
//     * bundle from apk only
//     * @param location
//     * @return
//     */
//    public Bundle installSync(final String[] location){
//        try {
//            mLocation = location;
//            return installBundleInternal(true);
//        }catch(Throwable e){
//            e.printStackTrace();
//            BundleInstallerFetcher.recycle(this);
//            return null;
//        }
//    }
    /////////////////////////////////////安装外部bundle////////////////////////////////////

    private  void checkSrc(String[] location,Object[] input){
        if(location==null || location.length==0){
            throw new RuntimeException("bundle name can not be empty");
        }
        if(input == null || input.length==0){
            throw new RuntimeException("can not find raw bundle file");
        }
    }

    private void installBundleInternal(boolean sync) throws Exception{
        if(Thread.currentThread().getId() == sBundleHandler.getLooper().getThread().getId()){
            Log.e("BundleInstaller", Arrays.toString(mLocation));
            call();
            if(mListener!=null){
                mListener.onFinished();
            }
            BundleInstallerFetcher.recycle(this);
        }else{
            if(sync){
                if(Thread.currentThread().getId() == Looper.getMainLooper().getThread().getId()){
                    call();
                    BundleInstallerFetcher.recycle(this);
                }else {
                    synchronized (this) {
                        deliveryTask(sync);
                        Log.d("BundleInstaller", "call wait:" + this);
                        this.wait(20000);
                        BundleInstallerFetcher.recycle(this);
                    }
                }
            }else{
                deliveryTask(sync);
            }
        }
    }

    private void deliveryTask(boolean sync){
        Runnable installTask = new Runnable() {
            @Override
            public void run() {
                synchronized (BundleInstaller.this) {
                    try {
                        if(mLocation==null){
                            return;
                        }
                        call();
                    } catch (Throwable e) {
                        e.printStackTrace();
                    } finally {
                        BundleInstaller.this.notify();
                        if (mListener != null) {
                            new Handler(Looper.getMainLooper()).post(new Runnable() {
                                @Override
                                public void run() {
                                    Log.e("BundleInstaller", "notify finish "+ Arrays.toString(mLocation));
                                    if(mListener!=null) {
                                        mListener.onFinished();
                                    }
                                    BundleInstallerFetcher.recycle(BundleInstaller.this);
                                }
                            });
                        }
                        try {
                            Thread.sleep(1);
                        } catch (InterruptedException e) {
                        }
                    }
                }
            }
        };
        sBundleHandler.post(installTask);

    }

    public Bundle getInstalledBundle(String bundleName){
        Bundle bundle = Framework.getBundle(bundleName);
        if(bundle==null){
            bundle = Framework.restoreFromExistedBundle(bundleName);
        }
        if(bundle==null && (BaselineInfoManager.instance().isDexPatched(bundleName) || BaselineInfoManager.instance().isUpdated(bundleName))){
            Log.e("BundleInstaller","restore existed bundle failed : "+bundleName);
        }
        return bundle;
    }

    @Override
    public synchronized Void call() throws Exception {
        Bundle bundle = null;
        if(!mTransitive) {
            for(int x=0; x<mLocation.length; x++) {
                if (FileUtils.getUsableSpace(Environment.getDataDirectory()) >= 5) {
                    if (mBundleSourceInputStream != null && mBundleSourceInputStream.length>x && mBundleSourceInputStream[x]!=null) {
                        if ((bundle = getInstalledBundle(mLocation[x])) == null) {
                            if((BaselineInfoManager.instance().isDexPatched(mLocation[x]) || BaselineInfoManager.instance().isUpdated(mLocation[x]))){
                                continue;
                            }
                            bundle = Framework.installNewBundle(mLocation[x], mBundleSourceInputStream[x]);
                        }
                        if (bundle != null) {
                            ((BundleImpl) bundle).optDexFile();
                        }
                    } else if (mBundleSourceFile != null && mBundleSourceFile.length>x && mBundleSourceFile[x]!=null) {
                        if ((bundle = getInstalledBundle(mLocation[x])) == null) {
                            if((BaselineInfoManager.instance().isDexPatched(mLocation[x]) || BaselineInfoManager.instance().isUpdated(mLocation[x]))){
                                continue;
                            }
                            bundle = Framework.installNewBundle(mLocation[x], mBundleSourceFile[x]);
                        }
                        if (bundle != null) {
                            ((BundleImpl) bundle).optDexFile();
                        }
                    } else {
                        if ((bundle = getInstalledBundle(mLocation[x])) == null && AtlasBundleInfoManager.instance().isInternalBundle(mLocation[x])) {
                            if((BaselineInfoManager.instance().isDexPatched(mLocation[x]) || BaselineInfoManager.instance().isUpdated(mLocation[x]))){
                                continue;
                            }
                            bundle = installBundleFromApk(mLocation[x]);
                            if (bundle != null) {
                                ((BundleImpl) bundle).optDexFile();
                            }
                        }
                    }
                } else {
                    throw new LowDiskException("no enough space");
                }
            }
        }else{
            for(int x=0; x<mLocation.length; x++) {
                List<String> bundlesForInstall = AtlasBundleInfoManager.instance().getBundleInfo(mLocation[x]).getTotalDependency();
                Log.e("BundleInstaller",mLocation[x]+"-->"+bundlesForInstall.toString());
                for (String bundleName : bundlesForInstall) {
                    if ((bundle = getInstalledBundle(bundleName)) == null) {
                        if((BaselineInfoManager.instance().isDexPatched(bundleName) || BaselineInfoManager.instance().isUpdated(bundleName))){
                            continue;
                        }
                        if (FileUtils.getUsableSpace(Environment.getDataDirectory()) >= 5) {
                            //has enough space
                            if(AtlasBundleInfoManager.instance().isInternalBundle(bundleName)) {
                                bundle = installBundleFromApk(bundleName);
                                if (bundle != null) {
                                    ((BundleImpl) bundle).optDexFile();
                                }
                            }
                        } else {
                            throw new LowDiskException("no enough space");
                        }
                    } else {
                        if (bundle!=null && ((BundleImpl) bundle).getArchive()!=null && !((BundleImpl) bundle).getArchive().isDexOpted()) {
                            ((BundleImpl) bundle).optDexFile();
                        }
                    }
                }
            }
        }
        return null;
    }

    /////////////////////////////util/////////////////
        public static synchronized void resolveInternalBundles() {

        }

    /**
     * 获取bundle源文件地址
     *
     * bundle finding order by follwing paths:
     * 1) ${getApplicationInfo().dataDir}/lib/
     * 2) ${getApplicationInfo().nativeLibraryDir}
     * 3) apk's "lib/armeabi" path.
     * @param location
     */
    private void findBundleSource(String location) throws IOException{
        mTmpBundleSourceInputStream = null;
        mTmpBundleSourceFile = null;
        String dataDir = RuntimeVariables.androidApplication.getApplicationInfo().dataDir;
        String bundleFileName = String.format("lib%s.so",location.replace(".","_"));
        String bundlePath = String.format("%s/lib/%s", dataDir,bundleFileName);
        File bundleFile = new File(bundlePath);
        if(!bundleFile.exists()){
            bundleFile = new File(RuntimeVariables.androidApplication.getApplicationInfo().nativeLibraryDir,bundleFileName);
        }
        if(isBundleFileTimeStampMatched(location,bundleFile)){
            mTmpBundleSourceFile = bundleFile;
            Log.e("BundleInstaller","find valid bundle : "+bundleFile.getAbsolutePath());
        }else{
            if(ApkUtils.getApk()!=null){
                ZipEntry entry = ApkUtils.getApk().getEntry("lib/armeabi/" + bundleFileName);
                if(entry!=null) {
                    mTmpBundleSourceInputStream = ApkUtils.getApk().getInputStream(entry);
                }
            }
        }
    }

    private boolean isBundleFileTimeStampMatched(String location,File file){
        if(!file.exists() || !AtlasBundleInfoManager.instance().isInternalBundle(location)){
            return false;
        }
        if(file.lastModified() == getTimeStampInApk()){
            return true;
        }
        return false;
    }

    private static long timeStampInApk = -11021836;
    private synchronized long getTimeStampInApk(){
        try {
            if (timeStampInApk == -11021836) {
                timeStampInApk = ApkUtils.getApk().getEntry("classes.dex").getTime();
            }
        }finally {
            return timeStampInApk>0 ? timeStampInApk : 0;
        }
    }

    private Bundle installBundleFromApk(String bundleName) throws Exception{
        Bundle bundle = null;
        findBundleSource(bundleName);
        if(mTmpBundleSourceFile!=null){
            bundle = Framework.installNewBundle(bundleName,mTmpBundleSourceFile);
        }else if(mTmpBundleSourceInputStream!=null){
            bundle = Framework.installNewBundle(bundleName,mTmpBundleSourceInputStream);
        }else{
            IOException e = new IOException("can not find bundle source file");
            Map<String, Object> detail = new HashMap<>();
            detail.put("installBundleFromApk",bundleName);
            AtlasMonitor.getInstance().report(AtlasMonitor.CONTAINER_BUNDLE_SOURCE_MISMATCH, detail, e);
            throw e;
        }
        return bundle;
    }

    ///////////////////////util//////////////////////////////////////

    public interface InstallListener {
        public void onFinished();
    }
}
