/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.framework;

import android.content.Context;
import android.content.pm.ApplicationInfo;
import android.os.Build;
import android.os.Environment;
import android.os.Looper;
import android.os.Process;
import android.taobao.atlas.bundleInfo.AtlasBundleInfoManager;
import android.taobao.atlas.bundleInfo.BundleListing;
import android.taobao.atlas.framework.bundlestorage.BundleArchive;
import android.taobao.atlas.runtime.ClassNotFoundInterceptorCallback;
import android.taobao.atlas.runtime.InstrumentationHook;
import android.taobao.atlas.runtime.LowDiskException;
import android.taobao.atlas.runtime.RuntimeVariables;
import android.taobao.atlas.util.*;
import android.taobao.atlas.util.AtlasFileLock;
import android.taobao.atlas.util.log.impl.AtlasMonitor;
import android.taobao.atlas.versionInfo.BaselineInfoManager;
import android.text.TextUtils;
import android.util.Log;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleException;
import org.osgi.framework.BundleListener;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.FrameworkListener;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import android.os.*;
import java.util.zip.ZipEntry;

import static android.os.Environment.MEDIA_UNKNOWN;

public final class Framework {

    public final static String DEPRECATED_MARK = "deprecated";
    static String containerVersion = "";
    private static ClassNotFoundInterceptorCallback classNotFoundCallback;

    public static String ATLAS_DEBUG_DIRECTORY;

    /**
     * framework basedir.
     */
    private static String BASEDIR;

    /**
     * the location where the storage resides.
     */
    public static final String STORAGE_LOCATION;

    /**
     * debug outputs from bundles ?
     */
    static boolean DEBUG_BUNDLES;

    /**
     * location -> bundle.
     */
    static Map<String, Bundle> bundles = new ConcurrentHashMap<String, Bundle>();

    /**
     * bundle listeners.
     */
    static List<BundleListener> bundleListeners = new ArrayList<BundleListener>();

    /**
     * synchronous bundle listeners.
     */
    static List<BundleListener> syncBundleListeners = new ArrayList<BundleListener>();

    /**
     * framework listeners.
     */
    static List<FrameworkListener> frameworkListeners = new ArrayList<FrameworkListener>();

    static HashMap<String,Integer> installingBundles = new HashMap<>();

    /**
     * system ClassLoader
     */
    static ClassLoader systemClassLoader;

    private static boolean bundleUpdated = false;
    public static boolean updateHappend = false;
    public static boolean DEBUG = false;

    static {
        File fileDir = RuntimeVariables.androidApplication.getFilesDir();
        if (fileDir == null || !fileDir.exists()) {
            fileDir = RuntimeVariables.androidApplication.getFilesDir();
        }
        try {
            ATLAS_DEBUG_DIRECTORY = RuntimeVariables.androidApplication.getExternalFilesDir("atlas-debug").getAbsolutePath();
        } catch (Exception e) {
            ATLAS_DEBUG_DIRECTORY = "/sdcard/Android/data/" + RuntimeVariables.androidApplication.getPackageName() + "/files/atlas-debug";
        }
        BASEDIR = fileDir.getAbsolutePath();
        STORAGE_LOCATION = BASEDIR + File.separatorChar + "storage" + File.separatorChar;

        try {
            ApplicationInfo app_info = RuntimeVariables.androidApplication.getApplicationInfo();
            DEBUG = (app_info.flags & ApplicationInfo.FLAG_DEBUGGABLE) != 0;
        } catch (final Exception e) {
            DEBUG = true;
        }
    }

    /**
     * Hide defautlt constructor
     */
    private Framework() {
    }

    /**
     * launch the framework.
     *
     * @throws Throwable
     */
    static void startup(boolean updated) throws BundleException {
        AtlasBundleInfoManager.instance().getBundleInfo();
        notifyFrameworkListeners(0 /* STARTING */, null, null);
        notifyFrameworkListeners(FrameworkEvent.STARTED, null, null);
    }

    public static ClassLoader getSystemClassLoader() {
        return systemClassLoader;
    }

    public static List<Bundle> getBundles() {
        final List<Bundle> res = new ArrayList<Bundle>(bundles.size());
        synchronized (bundles) {
            res.addAll(bundles.values());
        }
        return res;
    }

    public synchronized static Bundle getBundle(String location) {
        if (location == null){
            return null;
        }
        return bundles.get(location);
    }

    /**
     * delete a directory with all subdirs.
     *
     * @param path the directory.
     */
    public static void deleteDirectory(final File path) {
        final File[] files = path.listFiles();
        if (files == null){
            return;
        }
        Log.e("delete",path.getAbsolutePath());
        for (int i = 0; i < files.length; i++) {
            if (files[i].isDirectory()) {
                deleteDirectory(files[i]);
            } else {
                files[i].delete();
            }
        }
        path.delete();
    }

    /**
     * install a bundle from input stream.
     *
     * @param location the bundle location.
     * @param in       the input stream.
     * @return a Bundle object.
     * @throws BundleException if the installation failed.
     */
     static BundleImpl installNewBundle(final String location, final InputStream in)throws BundleException {
        File bundleDir = null;
        try{
            BundleLock.WriteLock(location);
            installingBundles.put(location,0);
            bundleDir = createBundleStorage(location);
            if(!bundleDir.exists()){
                bundleDir.mkdirs();
            }
            AtlasFileLock.getInstance().LockExclusive(bundleDir);
            final BundleImpl cached;
            if ((cached = (BundleImpl) getBundle(location)) != null) {
                return cached;
            }

            BundleImpl bundle = null;
            BundleListing.BundleInfo info = AtlasBundleInfoManager.instance().getBundleInfo(location);
            bundle = new BundleImpl(bundleDir, location, in, null, info.getUnique_tag(),true,-1);
            return bundle;
        } catch (IOException e) {
            BundleException e1 = new BundleException("Failed to install bundle." + FileUtils.getAvailableDisk(), e);
            if (bundleDir != null && !updateHappend) {
                Framework.deleteDirectory(bundleDir);
            }
            if (FileUtils.getUsableSpace(Environment.getDataDirectory()) < LowDiskException.thredshold) {
                throw new LowDiskException(FileUtils.getAvailableDisk(), e);
            }
            throw new BundleException("Failed to install bundle.", e);
        } catch (BundleException e) {
            BundleException e1 = new BundleException("Failed to install bundle." + FileUtils.getAvailableDisk(), e);
            if (bundleDir != null && !updateHappend) {
                Framework.deleteDirectory(bundleDir);
            }
            throw e1;
        } finally {
            installingBundles.remove(location);
            BundleLock.WriteUnLock(location);
            if (bundleDir != null) {
                AtlasFileLock.getInstance().unLock(bundleDir);
            }
        }
    }

    /**
     * install a bundle from input file, the file will move to bundle storage directory.
     *
     * @param location the bundle location.
     * @param file     the input file.
     * @return a Bundle object.
     * @throws BundleException if the installation failed.
     * @throws
     */
     static BundleImpl installNewBundle(final String location, final File file) throws BundleException {
        File bundleDir = null;

        try {
            BundleLock.WriteLock(location);
            installingBundles.put(location,0);
            bundleDir = createBundleStorage(location);
            if(!bundleDir.exists()){
                bundleDir.mkdirs();
            }
            AtlasFileLock.getInstance().LockExclusive(bundleDir);
            final BundleImpl cached;
            if ((cached = (BundleImpl) getBundle(location)) != null) {
                return cached;
            }
            Log.e("BundleInstaller","real install " + location);
            BundleImpl bundle = null;

            BundleListing.BundleInfo info = AtlasBundleInfoManager.instance().getBundleInfo(location);
            bundle = new BundleImpl(bundleDir, location, null, file,info.getUnique_tag(),true,-1);
            return bundle;
        } catch (IOException e) {
            BundleException e1 = new BundleException("Failed to install bundle." + FileUtils.getAvailableDisk(), e);
            if (bundleDir != null && !updateHappend) {
                Framework.deleteDirectory(bundleDir);
            }
            if (FileUtils.getUsableSpace(Environment.getDataDirectory()) < LowDiskException.thredshold) {
                throw new LowDiskException(FileUtils.getAvailableDisk(), e);
            }
            throw new BundleException("Failed to install bundle.", e);
        } catch (BundleException e) {
            BundleException e1 = new BundleException("Failed to install bundle." + FileUtils.getAvailableDisk(), e);
            if (bundleDir != null && !updateHappend) {
                Framework.deleteDirectory(bundleDir);
            }
            throw e1;
        } finally {
            installingBundles.remove(location);
            BundleLock.WriteUnLock(location);
            if (bundleDir != null) {
                AtlasFileLock.getInstance().unLock(bundleDir);
            }
        }
    }

    public static BundleImpl restoreFromExistedBundle(final String location) {
        boolean lockSuccess = false;
        BundleImpl bundle = null;
        String bundleUniqueTag = AtlasBundleInfoManager.instance().getBundleInfo(location).getUnique_tag();
        long dexPatchVersion = BaselineInfoManager.instance().getDexPatchBundleVersion(location);
        File bundleDir = null;
        File dexPatchDir = null;

        if(new File(STORAGE_LOCATION,location+File.separator+bundleUniqueTag).exists()){
            bundleDir = new File(STORAGE_LOCATION,location);
        }else{
            File[] externalStorages = getExternalFilesDirs(RuntimeVariables.androidApplication,"storage");
            if(externalStorages!=null && externalStorages.length>0){
                for(File tmpDir : externalStorages){
                    if(tmpDir!=null && getStorageState(tmpDir).equals(Environment.MEDIA_MOUNTED) && new File(tmpDir,location+File.separator+bundleUniqueTag).exists()) {
                        bundleDir = new File(tmpDir,location);
                        break;
                    }
                }
            }
        }

        if(dexPatchVersion>0){
            if(new File(STORAGE_LOCATION,location+File.separator+BundleArchive.DEXPATCH_DIR+dexPatchVersion).exists()){
                dexPatchDir = new File(STORAGE_LOCATION,location);
            }else{
                File[] externalStorages = getExternalFilesDirs(RuntimeVariables.androidApplication,"storage");
                if(externalStorages!=null && externalStorages.length>0){
                    for(File tmpDir : externalStorages){
                        if(tmpDir!=null && getStorageState(tmpDir).equals(Environment.MEDIA_MOUNTED) && new File(tmpDir,location+File.separator+BundleArchive.DEXPATCH_DIR+dexPatchVersion).exists()) {
                            bundleDir = new File(tmpDir,location);
                            break;
                        }
                    }
                }
            }
        }
        if(bundleDir!=null && bundleDir.exists()) {
            String[] temp = bundleDir.list();
            if(temp!=null) {
                for (String tmp : temp) {
                    Log.e("Framework", "content: " + tmp);
                }
            }
        }
        Log.e("Framework","restoreExisted: "+location+"| "+bundleUniqueTag +"| "+bundleDir);
        if((bundleDir==null || !new File(bundleDir,bundleUniqueTag).exists()) && BaselineInfoManager.instance().isUpdated(location)){
            //出现无法恢复的情况，立即回滚
            if(getCurProcessName().equals(RuntimeVariables.androidApplication.getPackageName())) {
                Map<String, Object> detail = new HashMap<>();
                detail.put("Bundle", location);
                AtlasMonitor.getInstance().report(AtlasMonitor.DD_BUNDLE_MISMATCH, detail, new RuntimeException(bundleUniqueTag+" is not existed"));
                BaselineInfoManager.instance().rollbackHardly();
                new Handler(Looper.getMainLooper()).postDelayed(new Runnable() {
                    @Override
                    public void run() {
                        android.os.Process.killProcess(Process.myPid());
                    }
                },1000);
            }
        }        
        if(!installingBundles.containsKey(location) && ((bundleDir!=null&&bundleDir.exists()) || dexPatchDir!=null&&dexPatchDir.exists())){
            try {
                lockSuccess = BundleLock.ReadLock(location);
                BundleContext bcontext = new BundleContext();
                bcontext.bundle_tag = bundleUniqueTag;
                bcontext.location = location;
                bcontext.bundleDir = bundleDir;
                bcontext.dexPatchDir = dexPatchDir;
                bundle = new BundleImpl(bcontext);
                if (bundle != null) {
                    bundle.optDexFile();
                }
            } catch (Exception e) {
                if (e instanceof BundleArchive.MisMatchException) {
                    if (bundleDir.exists()) {
                        bundle = null;
                    }
                }
                Log.e("Framework","restore bundle failed" + location, e);
            }finally {
                if(lockSuccess) {
                    try {
                        BundleLock.ReadUnLock(location);
                    }catch(Throwable e){}
                }
            }
        }
        return bundle;
    }

    // update or dexpatched
    public static void update(boolean upgrade,final String[] locations, final File[] files, String[] newBundleTag,long[] dexPatchVersions,String newBaselineVersion,boolean lowInternalDisk) throws BundleException {
        if (locations == null || files == null || locations.length != files.length) {
            throw new IllegalArgumentException("locations and files must not be null and must be same length");
        }
        Class KernalBundleClass = null;
        HashMap<String,String> updateBundles = new HashMap<>();
        File updateStorageDir = new File(STORAGE_LOCATION);
        if(lowInternalDisk){
            updateStorageDir = null;
            File[] externalStorages = getExternalFilesDirs(RuntimeVariables.androidApplication,"storage");
            if(externalStorages!=null && externalStorages.length>0){
                for(File externalStorage : externalStorages){
                    if(externalStorage!=null && getStorageState(externalStorage).equals(Environment.MEDIA_MOUNTED) && externalStorage.getUsableSpace()>50*1024*1024) {
                        updateStorageDir = externalStorage;
                    }
                }
            }
        }
        if(updateStorageDir==null){
            throw new BundleException("no enough space");
        }
        updateHappend = true;
        for (int i = 0; i < locations.length; i++) {
            if (locations[i] == null || files[i] == null) {
                continue;
            }
            //reset
            if(upgrade && newBundleTag[i].equals("-1")){
                continue;
            }else if(!upgrade && dexPatchVersions[i]==-1){
                updateBundles.put(locations[i],"-1");
            }

            File bundleDir = null;
            try {
                BundleLock.WriteLock(locations[i]);
                if (isKernalBundle(locations[i])) {
                    KernalBundleClass = RuntimeVariables.getRawClassLoader().loadClass("android.taobao.atlas.startup.patch.KernalBundle");
                    bundleDir = new File(updateStorageDir, "com.taobao.maindex");
                    if (!bundleDir.exists()){
                        bundleDir.mkdirs();
                    }
                    AtlasFileLock.getInstance().LockExclusive(bundleDir);
                    Constructor cons = KernalBundleClass.getDeclaredConstructor(File.class,File.class,String.class,long.class);
                    cons.setAccessible(true);
                    if(upgrade) {
                        cons.newInstance(bundleDir, files[i], makeMainDexUniqueTag(newBaselineVersion,newBundleTag[i]), -1);
                    }else{
                        cons.newInstance(bundleDir, files[i],null,dexPatchVersions[i]);                    }
                } else {
                    bundleDir = new File(updateStorageDir, locations[i]);
                    if (!bundleDir.exists()) {
                        bundleDir.mkdirs();
                    }
                    // Hold the storage file lock
                    AtlasFileLock.getInstance().LockExclusive(bundleDir);
                    if(upgrade) {
                        new BundleImpl(bundleDir, locations[i], null, files[i], newBundleTag[i], false, -1);
                    }else{
                        new BundleImpl(bundleDir, locations[i], null, files[i],null, false,dexPatchVersions[i]);
                    }
//                    Log.e("Framework","info of "+bundleDir+"newTag: "+newBundleTag[i]);
//                    String[] list = bundleDir.list();
//                    for(String temp : list){
//                        Log.e("Framework",temp);
//                    }
                }
                if(upgrade){
                    updateBundles.put(locations[i],newBundleTag[i]);
                }else{
                    updateBundles.put(locations[i],Long.toString(dexPatchVersions[i]));
                }
            } catch (Exception e) {
                if(upgrade) {
                    throw new BundleException("failed to installOrUpdate bundles ", e);
                }
            } finally {
                if (bundleDir != null) {
                    AtlasFileLock.getInstance().unLock(bundleDir);
                }
                BundleLock.WriteUnLock(locations[i]);
            }
        }
        if(upgrade) {
            bundleUpdated = true;
            try {
                BaselineInfoManager.instance().saveBaselineInfo(newBaselineVersion,updateBundles,lowInternalDisk ? updateStorageDir.getAbsolutePath() : "");
            } catch (IOException e) {
                throw new BundleException("save baseline info fail");
            }
            WrapperUtil.appendLog("installedVersionWhenUpdated", Framework.containerVersion);
            WrapperUtil.appendLog("VersionWhenUpdated", WrapperUtil.getPackageInfo(RuntimeVariables.androidApplication).versionName);
            InstrumentationHook.notifyAppUpdated();
        }else{
            if(updateBundles.size()>0){
                try {
                    BaselineInfoManager.instance().saveDexPathInfo(updateBundles,lowInternalDisk ? updateStorageDir.getAbsolutePath() : "");
                } catch (IOException e) {
                    throw new BundleException("save dexpatch info fail");
                }
            }
        }
    }

    private static String makeMainDexUniqueTag(String appVersion,String maindexTag){
        if(maindexTag.startsWith(appVersion)){
            return maindexTag;
        }
        return appVersion+"_"+maindexTag;
    }

    public static void rollback(){
        BaselineInfoManager.instance().rollback();
    }

    static boolean isKernalBundle(String location) {
        if (TextUtils.isEmpty(location)) {
            return false;
        }
        return location.equals("com.taobao.maindex");
    }

    /**
     * notify all bundle listeners.
     *
     * @param state the new state.
     * @param bundle the bundle.
     */
    static void notifyBundleListeners(final int state, final Bundle bundle) {
        if (syncBundleListeners.isEmpty() && bundleListeners.isEmpty()) {
            return;
        }

        final BundleEvent event = new BundleEvent(state, bundle);

        // inform the synchrounous bundle listeners first ...
        final BundleListener[] syncs = (BundleListener[]) syncBundleListeners.toArray(new BundleListener[syncBundleListeners.size()]);

        for (int i = 0; i < syncs.length; i++) {
            syncs[i].bundleChanged(event);
        }

        if (bundleListeners.isEmpty()) {
            return;
        }

        final BundleListener[] asyncs = (BundleListener[]) bundleListeners.toArray(new BundleListener[bundleListeners.size()]);
        for (int i = 0; i < asyncs.length; i++) {
            asyncs[i].bundleChanged(event);
        }
    }

    public static void addFrameworkListener(FrameworkListener listener) {
        frameworkListeners.add(listener);
    }

    public static void removeFrameworkListener(FrameworkListener listener) {
        frameworkListeners.remove(listener);
    }

    static void addBundleListener(BundleListener listener) {
        bundleListeners.add(listener);
    }

    static void removeBundleListener(BundleListener listener) {
        bundleListeners.remove(listener);
    }

    /**
     * notify all framework listeners.
     *
     * @param state the new state.
     * @param bundle the bundle.
     * @param throwable a throwable.
     */
    static void notifyFrameworkListeners(final int state, final Bundle bundle, final Throwable throwable) {

        if (frameworkListeners.isEmpty()) {
            return;
        }

        final FrameworkEvent event = new FrameworkEvent(state);

        final FrameworkListener[] listeners = frameworkListeners.toArray(new FrameworkListener[frameworkListeners.size()]);

        for (int i = 0; i < listeners.length; i++) {
            final FrameworkListener listener = listeners[i];

            listener.frameworkEvent(event);
        }
    }

    public static ClassNotFoundInterceptorCallback getClassNotFoundCallback() {
        return classNotFoundCallback;
    }

    public static void setClassNotFoundCallback(ClassNotFoundInterceptorCallback classNotFoundCallback) {
        Framework.classNotFoundCallback = classNotFoundCallback;
    }

    public static File getInstalledBundle(String location,String bundleUniqueId) {
        File bundleDir = null;
        if(new File(STORAGE_LOCATION,location+File.separator+bundleUniqueId).exists()){
            bundleDir = new File(STORAGE_LOCATION,location);
            File file = getInstalledBundleInternal(location,bundleUniqueId,bundleDir);
            if(file!=null){
                return file;
            }
        }else{
            File[] externalStorages = getExternalFilesDirs(RuntimeVariables.androidApplication,"storage");
            if(externalStorages!=null && externalStorages.length>0){
                for(File tmpDir : externalStorages){
                    if(tmpDir!=null && getStorageState(tmpDir).equals(Environment.MEDIA_MOUNTED) && new File(tmpDir,location+File.separator+bundleUniqueId).exists()) {
                        bundleDir = new File(tmpDir,location);
                        File file = getInstalledBundleInternal(location,bundleUniqueId,bundleDir);
                        if(file.exists()){
                            return file;
                        }
                    }
                }
            }
        }
        return null;
    }

    private static File getInstalledBundleInternal(String bundleName,String bundleUniqueId,File bundleDir){
        if(bundleDir!=null && bundleDir.exists()){
            if(isKernalBundle(bundleName)){
                File file = new File(bundleDir,makeMainDexUniqueTag(BaselineInfoManager.instance().currentVersionName(),bundleUniqueId)+File.separator+"com_taobao_maindex.zip");
                if(file.exists()){
                    return file;
                }
            }else{
                try {
                    BundleArchive archive = new BundleArchive(bundleName, bundleDir, bundleUniqueId, 0);
                    if(archive!=null){
                        File file = archive.getArchiveFile();
                        if(file!=null && file.exists()){
                            return file;
                        }
                    }
                }catch(Throwable e){
                    return null;
                }
            }
        }
        return null;
    }

    private static File createBundleStorage(String bundleName){
        if(new File(STORAGE_LOCATION).getUsableSpace()>10*1024*1024){
            return new File(STORAGE_LOCATION,bundleName);
        }else{
            File externalStorageDir = null;
            File[] externalStorages = getExternalFilesDirs(RuntimeVariables.androidApplication,"storage");
            if(externalStorages!=null && externalStorages.length>0){
                for(File tmpDir : externalStorages){
                    if(tmpDir!=null && getStorageState(tmpDir).equals(Environment.MEDIA_MOUNTED) && tmpDir.getUsableSpace()>20*1024*1024) {
                        externalStorageDir = tmpDir;
                        break;
                    }
                }
            }
            if(externalStorageDir!=null){
                return new File(externalStorageDir,bundleName);
            }
        }
        return new File(STORAGE_LOCATION,bundleName);
    }

    public static File[] getExternalFilesDirs(Context context, String type) {
        final int version = Build.VERSION.SDK_INT;
        if (version >= 19) {
            //返回结果可能存在null值
            return context.getExternalFilesDirs(type);
        } else {
            return new File[] { context.getExternalFilesDir(type) };
        }
    }

    public static String getStorageState(File path) {
        final int version = Build.VERSION.SDK_INT;
        if (version >= 19) {
            return Environment.getStorageState(path);
        }

        try {
            final String canonicalPath = path.getCanonicalPath();
            final String canonicalExternal = Environment.getExternalStorageDirectory()
                    .getCanonicalPath();

            if (canonicalPath.startsWith(canonicalExternal)) {
                return Environment.getExternalStorageState();
            }
        } catch (IOException e) {
        }

        return MEDIA_UNKNOWN;
    }

    public static String getCurProcessName(){
        return RuntimeVariables.sCurrentProcessName;
    }

    public static boolean isUpdated() {
        return bundleUpdated;
    }

    public static boolean isDeubgMode() {
        return DEBUG;
    }
}