/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.runtime;


import android.content.res.AssetManager;
import android.content.res.Configuration;
import android.content.res.Resources;
import android.content.res.XmlResourceParser;
import android.graphics.drawable.Drawable;
import android.os.Build;
import android.taobao.atlas.hack.AndroidHack;
import android.taobao.atlas.hack.AtlasHacks;
import android.taobao.atlas.util.log.impl.AtlasMonitor;
import android.text.TextUtils;
import android.util.DisplayMetrics;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.*;

/**
 * Created by guanjie on 15/12/9.
 */
public class DelegateResources extends Resources {

    public static final int BUNDLE_RES = 0;
    public static final int APK_RES    = 1;
    private static AssetManagerProcessor sAssetManagerProcessor;
    private static ResourcesProcessor    sResourcesProcessor;
    private static ResourceIdFetcher     sResourcesFetcher;
    private static ArrayList<String> sFailedAsssetPath = new ArrayList<String>();

    private static String sKernalPathPath = null;
    private static String sAssetsPatchDir = null;
    private Resources origin;


    /**
     * Create a new Resources object on top of an existing set of assets in an
     * AssetManager.
     *
     * @param assets  Previously created AssetManager.
     * @param res     previously resource
     */
    public DelegateResources(AssetManager assets, Resources res) {
        super(assets, res.getDisplayMetrics(), res.getConfiguration());
        origin = res;
    }

    @Override
    public XmlResourceParser getLayout(int id) throws NotFoundException {
        XmlResourceParser result = null;
        NotFoundException exception = null;
        try{
            result = super.getLayout(id);
        }catch(NotFoundException e){
            exception = e;
            if(origin!=null) {
                try {
                    result = origin.getLayout(id);
                } catch (Throwable e2) {
                }
            }

        }
        if(result==null && exception!=null){
            throw exception;
        }
        return result;
    }

    public Drawable getDrawable(int id, Theme theme) throws NotFoundException {
        Drawable result = null;
        NotFoundException exception = null;
        try{
            result = super.getDrawable(id,theme);
        }catch(NotFoundException e){
            exception = e;
            if(origin!=null) {
                try {
                    result = origin.getDrawable(id,theme);
                } catch (Throwable e2) {
                }
            }

        }
        if(result==null && exception!=null){
            throw exception;
        }
        return result;
    }

    public static void reset(){
        sKernalPathPath = null;
        sAssetsPatchDir = null;
    }

    public static void addBundleResources(String assetPath,String debugPath)  throws Exception{
        synchronized (DelegateResources.class) {
            if(debugPath!=null && !findResByAssetIndexDescending()){
                updateResources(RuntimeVariables.delegateResources, debugPath, BUNDLE_RES);

            }
            updateResources(RuntimeVariables.delegateResources, assetPath, BUNDLE_RES);
            if(debugPath!=null && findResByAssetIndexDescending()){
                updateResources(RuntimeVariables.delegateResources, debugPath, BUNDLE_RES);

            }
        }
    }

    public static void addApkpatchResources(String assetPath) throws Exception{
        AtlasHacks.defineAndVerify();
        sKernalPathPath = assetPath;
        synchronized (DelegateResources.class) {
            updateResources(RuntimeVariables.delegateResources, assetPath, APK_RES);
        }
    }

    public static String getCurrentAssetpathStr(AssetManager manager){
        if(sAssetManagerProcessor!=null){
            return sAssetManagerProcessor.getCurrentAssetPathStr(manager);
        }
        return "";
    }

    public static List<String> getCurrentAssetPath(AssetManager manager){
        if(sAssetManagerProcessor!=null){
            return sAssetManagerProcessor.getAssetPath(manager);
        }
        return new ArrayList<String>();
    }

    public static boolean checkAsset(String assetPath){
        if(assetPath!=null && !sFailedAsssetPath.contains(assetPath)){
            return true;
        }
        return false;
    }


    private static void updateResources(Resources res,String assetPath,int assertType) throws Exception{
        if(sAssetManagerProcessor==null){
            sAssetManagerProcessor = new AssetManagerProcessor();
        }
        AssetManager updatedAssetManager = sAssetManagerProcessor.updateAssetManager(res.getAssets(), assetPath, assertType);

        if(sResourcesProcessor==null){
            sResourcesProcessor = getResourceProcessor();
        }
        sResourcesProcessor.updateResources(updatedAssetManager);

        if(sResourcesFetcher==null){
            sResourcesFetcher = new ResourceIdFetcher();
        }
        sResourcesFetcher.addAssetForGetIdentifier(assetPath);
    }

    private static ResourcesProcessor getResourceProcessor(){
        if(RuntimeVariables.delegateResources.getClass().getName().equals("android.content.res.MiuiResources")){
            return new MiuiResourcesProcessor();
        }else{
            return new ResourcesProcessor();
        }

    }

    @Override
    public int getIdentifier(String name, String defType, String defPackage) {
        int id = super.getIdentifier(name, defType, defPackage);
        if(id!=0){
            return id;
        }else {
            return sResourcesFetcher.getIdentifier(name, defType, defPackage);
        }
    }

    /**
     * 是否已assetpatch索引降序的方式查找资源
     * @return
     */
    private static boolean findResByAssetIndexDescending(){
        if(Build.VERSION.SDK_INT>20){
            return false;
        }else {
            return true;
        }
    }

    private static class AssetManagerProcessor {

        private static HashMap<String,Boolean> sDefaultAssetPathList  ;
        static {
            try {
                AssetManager manager = AssetManager.class.newInstance();
                ArrayList<String> defaultPaths = getAssetPath(manager);
                if(defaultPaths!=null && defaultPaths.size()>0){
                    sDefaultAssetPathList = new HashMap<String,Boolean>();
                    for(String path : defaultPaths){
                        sDefaultAssetPathList.put(path,Boolean.FALSE);
                    }
                }
            }catch(Throwable e){
                sDefaultAssetPathList = new HashMap<String,Boolean>(0);
            }
        }
        private LinkedHashMap<String,Boolean> assetPathCache = null;
        private LinkedHashMap<String,Boolean> preAssetPathCache = null;

        public AssetManagerProcessor(){
            assetPathCache = new LinkedHashMap<String,Boolean>();
            preAssetPathCache = new LinkedHashMap<String,Boolean>();
            assetPathCache.put(RuntimeVariables.androidApplication.getApplicationInfo().sourceDir,Boolean.FALSE);
        }

        public AssetManager updateAssetManager(AssetManager manager,String newAssetPath,int assetType)throws Exception{
            AssetManager targetManager = null;
            if(assetType == BUNDLE_RES){
                boolean append = findResByAssetIndexDescending();
                targetManager = createNewAssetManager(manager,newAssetPath,append,assetType);
                updateAssetPathList(newAssetPath,append);
            }else{
                File newAssetsDir = new File(new File(newAssetPath).getParent(),"newAssets");
                if(newAssetsDir.exists() && new File(newAssetsDir,"assets").exists()){
                    sAssetsPatchDir = newAssetsDir.getAbsolutePath();
                }
                if(supportExpandAssetManager()){
                    if(findResByAssetIndexDescending()){
                        targetManager = updateAssetManagerWithAppend(manager,newAssetPath,assetType);
                        updateAssetPathList(newAssetPath,true);
                    }else{
                        targetManager = createNewAssetManager(manager,newAssetPath,false,assetType);
                        updateAssetPathList(newAssetPath,false);
                    }
                }else{
                    if(findResByAssetIndexDescending()){
                        targetManager = createNewAssetManager(manager,newAssetPath,true,assetType);
                        updateAssetPathList(newAssetPath,true);
                    }else{
                        targetManager = createNewAssetManager(manager,newAssetPath,false,assetType);
                        updateAssetPathList(newAssetPath,false);
                    }
                }

            }

            return targetManager;
        }


        public List<String> getCurrentAssetPath(AssetManager target){
            if(sAssetManagerProcessor==null){
                return new ArrayList<String>();
            }
            return sAssetManagerProcessor.getAssetPath(target);
        }

        public String getCurrentAssetPathStr(AssetManager target){
            if (target==null){
                return "";
            }
            List<String> currentList = getAssetPath(target);
            StringBuffer sb = new StringBuffer();
            sb.append("newDelegateResources [");
            if(currentList!=null) {
                for (String path : currentList) {
                    sb.append(path).append(",");
                }
            }
            sb.append("]");
            return sb.toString();
        }

        private AssetManager updateAssetManagerWithAppend(AssetManager manager,String newAssetPath,int type) throws Exception{
            appendAssetPath(manager, newAssetPath);
            /**
             * 追加主apk新的assets内容
             */
            if(type == APK_RES && sAssetsPatchDir!=null){
                appendAssetPath(manager,sAssetsPatchDir);
            }
            return manager;
        }

        /**
         *
         * @param srcManager
         * @param newAssetPath
         * @param append true : 表示往后追加 false: 表示往前插入
         * @return
         * @throws Exception
         */
        private AssetManager createNewAssetManager(AssetManager srcManager,String newAssetPath,boolean append,int type) throws Exception{
            AssetManager newAssetManager = AssetManager.class.newInstance();
            List<String> runtimeAdditionalAssets = new ArrayList<String>();
            List<String> currentPaths = getAssetPath(srcManager);
            for(String currentPath : currentPaths){
                if(!sDefaultAssetPathList.containsKey(currentPath) && !assetPathCache.containsKey(currentPath)
                        && !preAssetPathCache.containsKey(currentPath) && !currentPath.equals(newAssetPath)){
                    if(currentPath.toLowerCase().contains("webview")) {
                        runtimeAdditionalAssets.add(currentPath);
                    }
                }
            }
            sFailedAsssetPath.clear();
            if(!append){
                appendAssetPath(newAssetManager,newAssetPath);
            }
                //顺序添加,逆序加入assetmanager
            if(preAssetPathCache!=null){
                if(preAssetPathCache.size()==1){
                    Iterator<Map.Entry<String, Boolean>> iterator = preAssetPathCache.entrySet().iterator();
                    appendAssetPath(newAssetManager, iterator.next().getKey());
                }else {
                    ListIterator<Map.Entry<String,Boolean>> i=new ArrayList<Map.Entry<String,Boolean>>(preAssetPathCache.entrySet()).listIterator(preAssetPathCache.size());
                    while(i.hasPrevious()) {
                        Map.Entry<String, Boolean> entry=i.previous();
                        appendAssetPath(newAssetManager, entry.getKey());
                    }
                }
            }

            if(assetPathCache!=null){
                Iterator<Map.Entry<String,Boolean>> iterator= assetPathCache.entrySet().iterator();
                while(iterator.hasNext()){
                    Map.Entry<String,Boolean> entry = iterator.next();
                    if(!sDefaultAssetPathList.containsKey(entry.getKey())) {
                        appendAssetPath(newAssetManager,entry.getKey());
                    }
                }
            }

            if(append){
                appendAssetPath(newAssetManager,newAssetPath);
            }

            //add additional assets
            if(!runtimeAdditionalAssets.isEmpty()){
                for(String additional : runtimeAdditionalAssets){
                    appendAssetPath(newAssetManager,additional);
                }
            }

            /**
             * 追加主apk新的assets内容
             */
            if(sAssetsPatchDir!=null){
                appendAssetPath(newAssetManager,sAssetsPatchDir);
            }


            return newAssetManager;
        }

        private boolean hasCreatedAssetsManager = false;
        private synchronized boolean supportExpandAssetManager(){
            if(Build.VERSION.SDK_INT>=24){
                return true;
            }else if(!hasCreatedAssetsManager || Build.VERSION.SDK_INT<=20 ||
                    Build.BRAND.equalsIgnoreCase("sony") || Build.BRAND.equalsIgnoreCase("semc")){
                hasCreatedAssetsManager = true;
                return false;
            }else{
                return true;
            }
        }

        private void updateAssetPathList(String newAssetPath,boolean append){
            if(append) {
                assetPathCache.put(newAssetPath,Boolean.FALSE);
            }else{
                //顺序添加,逆序加入assetmanager
                preAssetPathCache.put(newAssetPath,Boolean.FALSE);
            }

            if(sKernalPathPath!=null) {
                if(sFailedAsssetPath.contains(sKernalPathPath)){
                    throw new RuntimeException("maindex arsc inject fail");
                }
                if(sAssetsPatchDir!=null) {
                    if(sFailedAsssetPath.contains(sAssetsPatchDir)){
                        throw new RuntimeException("maindex assets inject fail");
                    }
                }
            }

        }

        private boolean appendAssetPath(AssetManager asset,String path) throws Exception{
            int cookie = 0;
            boolean result = false;
            try{
                if ((cookie = Integer.parseInt(AtlasHacks.AssetManager_addAssetPath.invoke(asset, path).toString())) == 0){
                    // if failed try three times
                    int i = 0;
                    for (i = 0; i < 3; i++){
                        if ((cookie=Integer.parseInt(AtlasHacks.AssetManager_addAssetPath.invoke(asset, path).toString())) != 0){
                            break;
                        }
                    }
                }
                if(cookie==0){
                    sFailedAsssetPath.add(path);
                } else {
                    result = true;
                }
            }catch(NumberFormatException e){}

            if (!result) {
//                AtlasMonitor.getInstance().trace(AtlasMonitor.CONTAINER_APPEND_ASSETPATH_FAIL,
//                        false, "0", "","");
                Map<String, Object> detail = new HashMap<>();
                detail.put("appendAssetPath", path);
                AtlasMonitor.getInstance().report(AtlasMonitor.CONTAINER_APPEND_ASSETPATH_FAIL, detail, new RuntimeException());
            }
            return result;
        }

        public static ArrayList<String> getAssetPath(AssetManager manager){
            ArrayList<String> assetPaths = new ArrayList<String>();
            try {
                Method method = manager.getClass().getDeclaredMethod("getStringBlockCount");
                method.setAccessible(true);
                int assetsPathCount = (Integer) method.invoke(manager);
                for (int x = 0; x < assetsPathCount; x++) {
                    // Cookies map to string blocks starting at 1
                    String assetsPath = (String) manager.getClass().getMethod("getCookieName", int.class).invoke(manager, x + 1);
                    if ((sDefaultAssetPathList== null)||(!TextUtils.isEmpty(assetsPath) && !sDefaultAssetPathList.containsKey(assetsPath))) {
                        assetPaths.add(assetsPath);
                    }
                }
            } catch (Throwable e) {
                e.printStackTrace();
            }
            return assetPaths;
        }
    }


    /**
     * process resources with new AssetManager
     */
    public static class ResourcesProcessor{

        public void updateResources(AssetManager assetManager) throws Exception{
            if(RuntimeVariables.delegateResources.getAssets()!=assetManager ||
                    RuntimeVariables.delegateResources== null ||
                    !(RuntimeVariables.delegateResources instanceof DelegateResources)){
                RuntimeVariables.delegateResources = createNewResources(assetManager);
                AndroidHack.injectResources(RuntimeVariables.androidApplication, RuntimeVariables.delegateResources);

            }
        }

        Resources createNewResources(AssetManager manager) throws Exception{
            return new DelegateResources(manager, RuntimeVariables.delegateResources);
        }
    }

    public static class MiuiResourcesProcessor extends ResourcesProcessor{
        @Override
        Resources createNewResources(AssetManager manager) throws Exception  {
            if(Build.VERSION.SDK_INT<=20) {
                Class<?> miuiResClass = Class.forName("android.content.res.MiuiResources");
                Constructor<?> miuiCons = miuiResClass.getDeclaredConstructor(AssetManager.class,DisplayMetrics.class,Configuration.class);
                miuiCons.setAccessible(true);
                return  (Resources) miuiCons.newInstance(manager,RuntimeVariables.delegateResources.getDisplayMetrics(),
                        RuntimeVariables.delegateResources.getConfiguration());
            }else{
                return super.createNewResources(manager);
            }
        }
    }


}
