/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.dependency.parser.helper;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.android.build.gradle.internal.LoggerWrapper;
import com.android.build.gradle.internal.dependency.VariantDependencies;
import com.android.build.gradle.internal.ide.DependencyConvertUtils;
import com.android.builder.model.MavenCoordinates;
import com.android.utils.ILogger;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.taobao.android.builder.dependency.parser.AtlasDepHelper;
import com.taobao.android.builder.dependency.parser.CircleDependencyCheck;
import com.taobao.android.builder.dependency.parser.DependencyLocationManager;
import com.taobao.android.builder.dependency.parser.ResolvedDependencyInfo;
import com.taobao.android.builder.tasks.incremental.ApDependencies;
import org.apache.commons.lang.StringUtils;
import org.gradle.api.Project;
import org.gradle.api.artifacts.ModuleVersionIdentifier;
import org.gradle.api.artifacts.ResolvedArtifact;
import org.gradle.api.artifacts.component.ComponentIdentifier;
import org.gradle.api.artifacts.component.ProjectComponentIdentifier;
import org.gradle.api.artifacts.result.DependencyResult;
import org.gradle.api.artifacts.result.ResolvedComponentResult;
import org.gradle.api.artifacts.result.ResolvedDependencyResult;

/**
 * Created by wuzhong on 2017/4/17.
 */
public class DependencyResolver {

    private static final ILogger logger = LoggerWrapper.getLogger(DependencyResolver.class);

    private final ApDependencies apDependencies;

    Project project;

    VariantDependencies variantDeps;

    Map<ModuleVersionIdentifier, List<ResolvedArtifact>> artifacts;

    Map<String, Set<String>> bundleProvidedMap;

    public Set<String> mainDependencies = new HashSet<>();

    public DependencyResolver(Project project, VariantDependencies variantDeps,
                              Map<ModuleVersionIdentifier, List<ResolvedArtifact>> artifacts,
                              Map<String, Set<String>> bundleProvidedMap, ApDependencies apDependencies) {
        this.project = project;
        this.variantDeps = variantDeps;
        this.artifacts = artifacts;
        this.bundleProvidedMap = bundleProvidedMap;
        this.apDependencies = apDependencies;
    }

    public List<ResolvedDependencyInfo> resolve(List<DependencyResult> dependencyResults, boolean mainBundle) {
        Multimap<String, ResolvedDependencyInfo> dependenciesMap = LinkedHashMultimap.create();
        // 不使用官方的扁平化的依赖处理，改用自己处理树状的依赖关系;对于application的依赖，我们只取compile的依赖
        Set<ModuleVersionIdentifier> directDependencies = new HashSet<ModuleVersionIdentifier>();
        Set<String> resolveSets = new HashSet<>();
        for (DependencyResult dependencyResult : dependencyResults) {
            if (dependencyResult instanceof ResolvedDependencyResult) {
                ModuleVersionIdentifier moduleVersion = ((ResolvedDependencyResult)dependencyResult).getSelected()
                    .getModuleVersion();
                CircleDependencyCheck circleDependencyCheck = new CircleDependencyCheck(moduleVersion);

                if (!directDependencies.contains(moduleVersion)) {
                    directDependencies.add(moduleVersion);
                    resolveDependency(null, ((ResolvedDependencyResult)dependencyResult).getSelected(), artifacts,
                                      variantDeps, 0, circleDependencyCheck,
                                      circleDependencyCheck.getRootDependencyNode(), dependenciesMap, resolveSets);
                }
            }
        }

        List<ResolvedDependencyInfo> mainResolvdInfo = resolveAllDependencies(dependenciesMap);
        if (mainBundle) {
            for (ResolvedDependencyInfo resolvedDependencyInfo : mainResolvdInfo) {
                addMainDependencyInfo(resolvedDependencyInfo);
            }
        }
        return mainResolvdInfo;

    }

    private void addMainDependencyInfo(ResolvedDependencyInfo resolvedDependencyInfo) {

        this.mainDependencies.add(resolvedDependencyInfo.getGroup() + ":" + resolvedDependencyInfo.getName());

        for (ResolvedDependencyInfo child : resolvedDependencyInfo.getChildren()) {
            addMainDependencyInfo(child);
        }

    }

    /**
     * 解析依赖
     *
     * @param parent
     * @param resolvedComponentResult
     * @param artifacts
     * @param configDependencies
     * @param indent
     */
    private void resolveDependency(ResolvedDependencyInfo parent, ResolvedComponentResult resolvedComponentResult,
                                   Map<ModuleVersionIdentifier, List<ResolvedArtifact>> artifacts,
                                   VariantDependencies configDependencies, int indent,
                                   CircleDependencyCheck circleDependencyCheck,
                                   CircleDependencyCheck.DependencyNode node,
                                   Multimap<String, ResolvedDependencyInfo> dependenciesMap,
                                   Set<String> resolvedDependencies) {
        ModuleVersionIdentifier moduleVersion = resolvedComponentResult.getModuleVersion();

        if (checkForExclusion(configDependencies, moduleVersion, resolvedComponentResult)) { return; }

        if (moduleVersion.getName().equals("support-annotations") && moduleVersion.getGroup().equals(
            "com.android.support")) {
            configDependencies.setAnnotationsPresent(true);
        }

        // now loop on all the artifact for this modules.
        List<ResolvedArtifact> moduleArtifacts = artifacts.get(moduleVersion);

        ComponentIdentifier id = resolvedComponentResult.getId();
        String gradlePath = (id instanceof ProjectComponentIdentifier) ? ((ProjectComponentIdentifier)id)
            .getProjectPath() : null;

        // 如果同时找到多个依赖，暂时没法判断是那个真正有用
        if (null != moduleArtifacts) {
            for (ResolvedArtifact resolvedArtifact : moduleArtifacts) {
                String key = moduleVersion.getGroup() + ":" + moduleVersion.getName();
                if (mainDependencies.contains(key)) {
                    continue;
                }
                if (resolvedDependencies.contains(key)) {
                    continue;
                }
                resolvedDependencies.add(key);
                boolean isAwbBundle = bundleProvidedMap.containsKey(key);
                Set<String> providedDirectDep = bundleProvidedMap.get(key);

                ResolvedDependencyInfo resolvedDependencyInfo = new ResolvedDependencyInfo(moduleVersion.getVersion(),
                                                                                           moduleVersion.getGroup(),
                                                                                           moduleVersion.getName(),
                                                                                           isAwbBundle ? "awb"
                                                                                               : resolvedArtifact
                                                                                                   .getType(),
                                                                                           resolvedArtifact
                                                                                               .getClassifier());
                resolvedDependencyInfo.setIndent(indent);
                resolvedDependencyInfo.setGradlePath(gradlePath);
                resolvedDependencyInfo.setResolvedArtifact(resolvedArtifact);

                String path = AtlasDepHelper.computeArtifactPath(moduleVersion, resolvedArtifact);
                String name = AtlasDepHelper.computeArtifactName(moduleVersion, resolvedArtifact);

                MavenCoordinates mavenCoordinates = DependencyConvertUtils.convert(resolvedArtifact);

                File explodedDir = DependencyLocationManager.getExploreDir(project, mavenCoordinates,
                                                                           resolvedArtifact.getFile(),
                                                                           resolvedArtifact.getType().toLowerCase(),
                                                                           path);

                resolvedDependencyInfo.setExplodedDir(explodedDir);

                resolvedDependencyInfo.setDependencyName(name);

                if (null == parent) {
                    parent = resolvedDependencyInfo;
                } else {
                    resolvedDependencyInfo.setParent(parent);
                    parent.getChildren().add(resolvedDependencyInfo);
                }

                Set<? extends DependencyResult> dependencies = resolvedComponentResult.getDependencies();
                if (null != dependencies) {
                    for (DependencyResult dep : dependencies) {

                        if (dep instanceof ResolvedDependencyResult) {
                            ResolvedComponentResult childResolvedComponentResult = ((ResolvedDependencyResult)dep)
                                .getSelected();

                            if (isAwbBundle && providedDirectDep.contains(
                                childResolvedComponentResult.getModuleVersion().getGroup() + ":"
                                + childResolvedComponentResult.getModuleVersion().getName())) {
                                continue;
                            }

                            CircleDependencyCheck.DependencyNode childNode = circleDependencyCheck.addDependency(
                                childResolvedComponentResult.getModuleVersion(), node, indent + 1);
                            CircleDependencyCheck.CircleResult circleResult = circleDependencyCheck.checkCircle(logger);
                            if (circleResult.hasCircle) {
                                logger.warning("[CircleDependency]" + StringUtils.join(circleResult.detail, ";"));
                            } else {
                                resolveDependency(parent, ((ResolvedDependencyResult)dep).getSelected(), artifacts,
                                                  configDependencies, indent + 1, circleDependencyCheck, childNode,
                                                  dependenciesMap, resolvedDependencies);
                            }
                        }
                    }
                }

                addDependencyInfo(resolvedDependencyInfo, null, dependenciesMap);
            }
        }

    }

    private boolean checkForExclusion(VariantDependencies configDependencies, ModuleVersionIdentifier moduleVersion,
                                      ResolvedComponentResult resolvedComponentResult) {
        if (configDependencies.getChecker().checkForExclusion(moduleVersion) || (apDependencies != null
                                                                                 && apDependencies
                                                                                     .hasSameResolvedDependency(
                                                                                         moduleVersion)
                                                                                 && !(resolvedComponentResult
            .getId() instanceof ProjectComponentIdentifier))) {
            return true;
        }
        return false;
    }

    /**
     * 增加dependency
     *
     * @param resolvedDependencyInfo
     * @param parent                 这个parent为一级依赖
     */
    private void addDependencyInfo(ResolvedDependencyInfo resolvedDependencyInfo, ResolvedDependencyInfo parent,
                                   Multimap<String, ResolvedDependencyInfo> dependenciesMap) {

        dependenciesMap.put(resolvedDependencyInfo.toString(), resolvedDependencyInfo);

        if (null != parent) {
            resolvedDependencyInfo.setParent(parent);
        }
        List<ResolvedDependencyInfo> children = resolvedDependencyInfo.getChildren();
        if (null != children && children.size() > 0) {

            for (ResolvedDependencyInfo child : children) {
                addDependencyInfo(child, resolvedDependencyInfo, dependenciesMap);
            }
        }
    }

    /**
     * 进行依赖仲裁
     */
    private List<ResolvedDependencyInfo> resolveAllDependencies(
        Multimap<String, ResolvedDependencyInfo> dependenciesMap) {

        List<ResolvedDependencyInfo> allResolvedDependencyInfos = new ArrayList<>();

        // 仲裁后的依赖关系.结构为父类-子类
        Multimap<ModuleVersionIdentifier, ResolvedDependencyInfo> resolvedDependenciesMap = LinkedHashMultimap.create();
        Map<ModuleVersionIdentifier, ResolvedDependencyInfo> directDependencies
            = new HashMap<ModuleVersionIdentifier, ResolvedDependencyInfo>();

        for (String key : dependenciesMap.keySet()) {
            Collection<ResolvedDependencyInfo> dependencyLevels = dependenciesMap.get(key);

            if (dependencyLevels.size() > 0) {

                List<ResolvedDependencyInfo> resolvedDependencyInfos = Lists.newArrayList();
                resolvedDependencyInfos.addAll(dependencyLevels);
                Collections.sort(resolvedDependencyInfos);
                ResolvedDependencyInfo resolvedDependencyInfo = resolvedDependencyInfos.get(0);
                ResolvedDependencyInfo parent = resolvedDependencyInfo.getParent();
                //对于放入resolvedDependenciesMap中的信息不需要加入children
                resolvedDependencyInfo.setChildren(Lists.<ResolvedDependencyInfo>newArrayList());
                if (null != parent) {
                    //如果存在的父依赖,就把当前依赖加入到父依赖的子依赖中
                    resolvedDependenciesMap.put(parent.getModuleVersionIdentifier(), resolvedDependencyInfo);
                } else {
                    //如果没有父依赖,就是一级依赖
                    directDependencies.put(resolvedDependencyInfo.getModuleVersionIdentifier(), resolvedDependencyInfo);
                }
            }
        }

        // 开始构建依赖树

        for (ModuleVersionIdentifier key : directDependencies.keySet()) {
            ResolvedDependencyInfo resolvedDependencyInfo = directDependencies.get(key);
            //清空children
            resolvedDependencyInfo.setChildren(Lists.<ResolvedDependencyInfo>newArrayList());
            addResolvedDependencyInfo(resolvedDependencyInfo, resolvedDependenciesMap);
            allResolvedDependencyInfos.add(resolvedDependencyInfo);
        }

        return allResolvedDependencyInfos;

    }

    /**
     * 通过递归的方式进行依赖的解析
     *
     * @param parentDependency
     * @param resolvedDependenciesMap
     */
    private void addResolvedDependencyInfo(ResolvedDependencyInfo parentDependency,
                                           Multimap<ModuleVersionIdentifier, ResolvedDependencyInfo>
                                               resolvedDependenciesMap) {
        int indent = parentDependency.getIndent();
        ModuleVersionIdentifier identifier = parentDependency.getModuleVersionIdentifier();
        Collection<ResolvedDependencyInfo> childDependencies = resolvedDependenciesMap.get(identifier);

        //TODO here
        for (ResolvedDependencyInfo childDependency : childDependencies) {
            if (childDependency.getIndent() > indent) {
                //                System.out.println(parentDependency + " indent " + indent + "->" + childDependency
                // +  " indent " + childDependency.getIndent());
                parentDependency.getChildren().add(childDependency);
                if (childDependency.getIndent() <= 1) {
                    addResolvedDependencyInfo(childDependency, resolvedDependenciesMap);
                }
            }
        }
    }

}
