/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.extension;

import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.google.common.collect.Sets;
import com.taobao.android.builder.extension.annotation.Config;

/**
 * Created by shenghua.nish on 2016-05-17 上午10:15.
 */
public class TBuildConfig {

    @Config(message = "远程bundle清单, artifactId", advance = false, order = 1, group = "atlas")
    private Set<String> outOfApkBundles = Sets.newHashSet();

    @Config(title = "自启动的bundle列表", message = "值是 packageName", order = 1, advance = false, group = "atlas")
    private List<String> autoStartBundles = new ArrayList<String>();

    @Config(title = "提前启动列表",
        message = "实现PreLaunch的类，多个类用 , 号分开", order = 1, advance = false, group = "atlas")
    private String preLaunch = "";

    @Config(title = "atlas 分包",
        message = "atlas的主dex分包机制，第一个dex只放atlas对应的启动代码", order = 3, advance = false, group = "atlas")
    private boolean atlasMultiDex = false;

    @Config(message = "需要删除的so文件列表", order = 4, advance = true, group = "atlas")
    private Set<String> removeSoFiles = Sets.newHashSet();

    @Config(title = "bundle的packageId定义文件", message = "bundle的packageId定义文件，不定义会自动分配", group = "atlas")
    private File packageIdFile = new File("");

    @Config(message = "自动生成bundle的packageId", order = 6, advance = false, group = "atlas")
    private boolean autoPackageId = true;

    @Config(message = "自动分配的packageId 最小值", order = 6, advance = false, group = "atlas")
    private int minPackageId = 35;

    @Config(title = "构建基线包", message = "构建基线包，建议开启，否则后面的patch包无法进行", order = 0, group = "atlas_patch")
    private Boolean createAP = true;

    @Config(message = "合并bundle jar中的资源文件", order = 8, advance = true, group = "atlas")
    private Boolean mergeAwbJavaRes = false;

    @Config(message = "是否依赖冲突终止打包", order = 0, group = "check")
    private boolean abortIfDependencyConflict = false;

    @Config(message = "是否类冲突终止打包", order = 0, group = "check")
    private boolean abortIfClassConflict = false;

    @Config(title = "预处理manifest", message = "如果开启atlas，必须为true", order = 7, advance = true, group = "atlas")
    private Boolean preProcessManifest = true;

    @Config(title = "使用自定义的aapt", message = "如果开启atlas，必须为true", order = 8, advance = true, group = "atlas")
    private Boolean useCustomAapt = false;

    @Config(title = "aapt输出的R为常量", message = "建议值设置为false， 可以减少动态部署的patch包大小", order = 9, advance = true,
        group = "atlas")
    private Boolean aaptConstantId = true;

    @Config(message = "注入核心的bundle信息", advance = true, order = 10, group = "atlas")
    private Boolean classInject = true;

    @Config(title = "主dex插桩", message = "老版本的主dex动态部署，已经废弃", advance = true, order = 11, group = "atlas")
    private Boolean doPreverify = false;

    @Deprecated
    private Boolean resV4Enabled = true;

    @Config(message = "class注入在proguard之前", advance = true, order = 12, group = "atlas")
    private Boolean injectBeforeProguard = false;

    @Config(title = "使用databinding的bundle列表", message = "需要进行databinding的bundle， 值为 packageName ", order = 13,
        advance = true, group = "atlas")
    private Set<String> dataBindingBundles = new HashSet<>();

    @Config(message = "proguard是否需要读取bundle中的混淆配置", order = 14, advance = true, group = "atlas")
    private boolean bundleProguardConfigEnabled = true;

    @Config(message = "依赖中的混淆是否只读取keep规则", order = 15, advance = true, group = "atlas")
    private boolean libraryProguardKeepOnly = true;

    @Config(title = "proguard配置读取依赖黑名单", message = "group:name,group2:name2", order = 16, advance = true,
        group = "atlas")
    private Set<String> bundleProguardConfigBlackList = new HashSet<>();

    @Deprecated
    private Set<String> insideOfApkBundles = Sets.newHashSet();

    private boolean incremental = false;

    public Set<String> getRemoveSoFiles() {
        return removeSoFiles;
    }

    public void setRemoveSoFiles(Set<String> removeSoFiles) {
        this.removeSoFiles = removeSoFiles;
    }

    public File getPackageIdFile() {
        return packageIdFile;
    }

    public void setPackageIdFile(File packageIdFile) {
        this.packageIdFile = packageIdFile;
    }

    public boolean isAutoPackageId() {
        return autoPackageId;
    }

    public void setAutoPackageId(boolean autoPackageId) {
        this.autoPackageId = autoPackageId;
    }

    public Boolean getPreProcessManifest() {
        return preProcessManifest;
    }

    public void setPreProcessManifest(Boolean preProcessManifest) {
        this.preProcessManifest = preProcessManifest;
    }

    public Boolean getUseCustomAapt() {
        return useCustomAapt;
    }

    public void setUseCustomAapt(Boolean useCustomAapt) {
        this.useCustomAapt = useCustomAapt;
    }

    public Boolean getAaptConstantId() {
        return aaptConstantId;
    }

    public void setAaptConstantId(Boolean aaptConstantId) {
        this.aaptConstantId = aaptConstantId;
    }

    public Boolean getClassInject() {
        return classInject;
    }

    public void setClassInject(Boolean classInject) {
        this.classInject = classInject;
    }

    public Boolean getInjectBeforeProguard() {
        return injectBeforeProguard;
    }

    public void setInjectBeforeProguard(Boolean injectBeforeProguard) {
        this.injectBeforeProguard = injectBeforeProguard;
    }

    public Boolean getCreateAP() {
        return createAP;
    }

    public Boolean getMergeAwbJavaRes() {
        return mergeAwbJavaRes;
    }

    public void setMergeAwbJavaRes(Boolean mergeAwbJavaRes) {
        this.mergeAwbJavaRes = mergeAwbJavaRes;
    }

    public Set<String> getOutOfApkBundles() {
        return outOfApkBundles;
    }

    public void setOutOfApkBundles(Set<String> outOfApkBundles) {
        this.outOfApkBundles = outOfApkBundles;
    }

    public Set<String> getInsideOfApkBundles() {
        return insideOfApkBundles;
    }

    public void setInsideOfApkBundles(Set<String> insideOfApkBundles) {
        this.insideOfApkBundles = insideOfApkBundles;
    }

    public List<String> getAutoStartBundles() {
        return autoStartBundles;
    }

    public void setAutoStartBundles(List<String> autoStartBundles) {
        this.autoStartBundles = autoStartBundles;
    }

    public String getPreLaunch() {
        return preLaunch;
    }

    public void setPreLaunch(String preLaunch) {
        this.preLaunch = preLaunch;
    }

    public Boolean getDoPreverify() {
        return doPreverify;
    }

    public void setDoPreverify(Boolean doPreverify) {
        this.doPreverify = doPreverify;
    }

    public Boolean getResV4Enabled() {
        return resV4Enabled;
    }

    public void setResV4Enabled(Boolean resV4Enabled) {
        this.resV4Enabled = resV4Enabled;
    }

    public boolean isAbortIfDependencyConflict() {
        return abortIfDependencyConflict;
    }

    public void setAbortIfDependencyConflict(boolean abortIfDependencyConflict) {
        this.abortIfDependencyConflict = abortIfDependencyConflict;
    }

    public boolean isAbortIfClassConflict() {
        return abortIfClassConflict;
    }

    public void setAbortIfClassConflict(boolean abortIfClassConflict) {
        this.abortIfClassConflict = abortIfClassConflict;
    }

    public Set<String> getDataBindingBundles() {
        return dataBindingBundles;
    }

    public void setDataBindingBundles(Set<String> dataBindingBundles) {
        this.dataBindingBundles = dataBindingBundles;
    }

    public boolean isAtlasMultiDex() {
        return atlasMultiDex;
    }

    public void setAtlasMultiDex(boolean atlasMultiDex) {
        this.atlasMultiDex = atlasMultiDex;
    }

    public Boolean isCreateAP() {
        return createAP;
    }

    public void setCreateAP(Boolean createAP) {
        this.createAP = createAP;
    }

    public boolean isIncremental() {
        return incremental;
    }

    public void setIncremental(boolean incremental) {
        this.incremental = incremental;
    }

    public boolean isBundleProguardConfigEnabled() {
        return bundleProguardConfigEnabled;
    }

    public void setBundleProguardConfigEnabled(boolean bundleProguardConfigEnabled) {
        this.bundleProguardConfigEnabled = bundleProguardConfigEnabled;
    }

    public boolean isLibraryProguardKeepOnly() {
        return libraryProguardKeepOnly;
    }

    public void setLibraryProguardKeepOnly(boolean libraryProguardKeepOnly) {
        this.libraryProguardKeepOnly = libraryProguardKeepOnly;
    }

    public Set<String> getBundleProguardConfigBlackList() {
        return bundleProguardConfigBlackList;
    }

    public void setBundleProguardConfigBlackList(Set<String> bundleProguardConfigBlackList) {
        this.bundleProguardConfigBlackList = bundleProguardConfigBlackList;
    }

    public int getMinPackageId() {
        return minPackageId;
    }

    public void setMinPackageId(int minPackageId) {
        this.minPackageId = minPackageId;
    }
}
