/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.manager;

import com.android.build.api.transform.QualifiedContent;
import com.android.build.gradle.AndroidGradleOptions;
import com.android.build.gradle.AppExtension;
import com.android.build.gradle.api.ApplicationVariant;
import com.android.build.gradle.internal.TaskContainerAdaptor;
import com.android.build.gradle.internal.TaskFactory;
import com.android.build.gradle.internal.api.AppVariantContext;
import com.android.build.gradle.internal.api.AppVariantOutputContext;
import com.android.build.gradle.internal.core.GradleVariantConfiguration;
import com.android.build.gradle.internal.incremental.InstantRunPatchingPolicy;
import com.android.build.gradle.internal.pipeline.AtlasExtendedContentType;
import com.android.build.gradle.internal.pipeline.ExtendedContentType;
import com.android.build.gradle.internal.pipeline.OriginalStream;
import com.android.build.gradle.internal.pipeline.OriginalStream.Builder;
import com.android.build.gradle.internal.pipeline.TransformTask;
import com.android.build.gradle.internal.scope.AndroidTask;
import com.android.build.gradle.internal.scope.PackagingScope;
import com.android.build.gradle.internal.scope.VariantOutputScope;
import com.android.build.gradle.internal.scope.VariantScope;
import com.android.build.gradle.internal.tasks.PrepareDependenciesTask;
import com.android.build.gradle.internal.transforms.DexTransform;
import com.android.build.gradle.internal.transforms.MultiDexTransform;
import com.android.build.gradle.internal.transforms.ProGuardTransform;
import com.android.build.gradle.internal.variant.BaseVariantOutputData;
import com.android.build.gradle.tasks.*;
import com.android.builder.core.AtlasBuilder;
import com.android.builder.core.DefaultDexOptions;
import com.google.common.base.Supplier;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.collect.Iterators;
import com.taobao.android.builder.AtlasBuildContext;
import com.taobao.android.builder.dependency.AtlasDependencyTree;
import com.taobao.android.builder.dependency.model.AwbBundle;
import com.taobao.android.builder.extension.AtlasExtension;
import com.taobao.android.builder.tasks.PrepareAPTask;
import com.taobao.android.builder.tasks.app.ApBuildTask;
import com.taobao.android.builder.tasks.app.GenerateAtlasSourceTask;
import com.taobao.android.builder.tasks.app.LogDependenciesTask;
import com.taobao.android.builder.tasks.app.bundle.AwbPackagingScope;
import com.taobao.android.builder.tasks.app.bundle.Dex;
import com.taobao.android.builder.tasks.app.bundle.JavacAwbsTask;
import com.taobao.android.builder.tasks.app.bundle.PackageAwb;
import com.taobao.android.builder.tasks.app.bundle.PackageAwbsTask;
import com.taobao.android.builder.tasks.app.bundle.ProcessResAwbsTask;
import com.taobao.android.builder.tasks.app.databinding.AwbDataBindingExportBuildInfoTask;
import com.taobao.android.builder.tasks.app.databinding.AwbDataBindingMergeArtifactsTask;
import com.taobao.android.builder.tasks.app.databinding.AwbDataBindingProcessLayoutTask;
import com.taobao.android.builder.tasks.app.databinding.AwbDataBindingRenameTask;
import com.taobao.android.builder.tasks.app.manifest.StandardizeLibManifestTask;
import com.taobao.android.builder.tasks.app.merge.MergeAssetAwbsConfigAction;
import com.taobao.android.builder.tasks.app.merge.MergeManifestAwbsConfigAction;
import com.taobao.android.builder.tasks.app.merge.MergeResAwbsConfigAction;
import com.taobao.android.builder.tasks.app.merge.MergeSoLibTask;
import com.taobao.android.builder.tasks.app.prepare.PrepareAaptTask;
import com.taobao.android.builder.tasks.app.prepare.PrepareAllDependenciesTask;
import com.taobao.android.builder.tasks.app.prepare.PrepareBundleInfoTask;
import com.taobao.android.builder.tasks.app.prepare.PreparePackageIdsTask;
import com.taobao.android.builder.tasks.incremental.AwoInstallTask;
import com.taobao.android.builder.tasks.incremental.PrepareBaseApkTask;
import com.taobao.android.builder.tasks.manager.MtlTaskContext;
import com.taobao.android.builder.tasks.manager.MtlTaskInjector;
import com.taobao.android.builder.tasks.manager.TaskQueryHelper;
import com.taobao.android.builder.tasks.manager.transform.MtlTransformContext;
import com.taobao.android.builder.tasks.manager.transform.MtlTransformInjector;
import com.taobao.android.builder.tasks.manager.transform.TransformManager;
import com.taobao.android.builder.tasks.tpatch.DiffBundleInfoTask;
import com.taobao.android.builder.tasks.tpatch.TPatchDiffApkBuildTask;
import com.taobao.android.builder.tasks.tpatch.TPatchDiffResAPBuildTask;
import com.taobao.android.builder.tasks.tpatch.TPatchTask;
import com.taobao.android.builder.tasks.transform.AtlasMultiDexTransform;
import com.taobao.android.builder.tasks.transform.AtlasProguardTransform;
import com.taobao.android.builder.tasks.transform.ClassInjectTransform;
import com.taobao.android.builder.tools.ReflectUtils;
import com.taobao.android.builder.tools.multidex.FastMultiDexer;
import org.gradle.api.Action;
import org.gradle.api.GradleException;
import org.gradle.api.Project;
import org.gradle.api.Task;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.function.Consumer;

/**
 * MTL插件编译apk的任务管理
 * Created by shenghua.nish on 2016-05-09 下午3:55.
 */
@SuppressWarnings("ALL")
public class AtlasAppTaskManager extends AtlasBaseTaskManager {

    private AppExtension appExtension;

    public AtlasAppTaskManager(AtlasBuilder androidBuilder, AppExtension appExtension, Project project,
                               AtlasExtension atlasExtension) {
        super(androidBuilder, appExtension, project, atlasExtension);
        this.appExtension = appExtension;
    }

    @Override
    public void runTask() {

        appExtension.getApplicationVariants().forEach(new Consumer<ApplicationVariant>() {

            @Override
            public void accept(ApplicationVariant applicationVariant) {


                AppVariantContext appVariantContext = AtlasBuildContext.sBuilderAdapter.appVariantContextFactory
                    .getAppVariantContext(project, applicationVariant);

                //if (appVariantContext.getVariantData().getScope().getInstantRunBuildContext().isInInstantRunMode()) {
                //    throw new GradleException(
                //        "atlas plgin is not compatible with instant run， plese turn it off in your ide！");
                //}
//                new AwbProguradHook().hookProguardTask(appVariantContext);

                List<MtlTaskContext> mtlTaskContextList = new ArrayList<MtlTaskContext>();

                mtlTaskContextList.add(new MtlTaskContext(appVariantContext.getVariantData().preBuildTask));

                mtlTaskContextList.add(new MtlTaskContext(LogDependenciesTask.ConfigAction.class, null));
                mtlTaskContextList.add(new MtlTaskContext(PrepareAPTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(PrepareDependenciesTask.class));

                mtlTaskContextList.add(new MtlTaskContext(PrepareAllDependenciesTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(appVariantContext.getVariantData().mergeAssetsTask));

                mtlTaskContextList.add(new MtlTaskContext(RenderscriptCompile.class));

                mtlTaskContextList.add(new MtlTaskContext(StandardizeLibManifestTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(PrepareBundleInfoTask.ConfigAction.class, null));

                if (!atlasExtension.getTBuildConfig().getClassInject()) {
                    mtlTaskContextList.add(new MtlTaskContext(GenerateAtlasSourceTask.ConfigAction.class, null));
                }

                mtlTaskContextList.add(new MtlTaskContext(PreparePackageIdsTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(MergeSoLibTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(PrepareAaptTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(AidlCompile.class));

                mtlTaskContextList.add(new MtlTaskContext(GenerateBuildConfig.class));

                mtlTaskContextList.add(new MtlTaskContext(MergeResAwbsConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(MergeAssetAwbsConfigAction.class, null));

                if (null != androidExtension.getDataBinding() && androidExtension.getDataBinding().isEnabled()) {
                    mtlTaskContextList.add(
                        new MtlTaskContext(AwbDataBindingProcessLayoutTask.ConfigAction.class, null));
                    mtlTaskContextList.add(
                        new MtlTaskContext(AwbDataBindingExportBuildInfoTask.ConfigAction.class, null));
                    mtlTaskContextList.add(
                        new MtlTaskContext(AwbDataBindingMergeArtifactsTask.ConfigAction.class, null));
                }

                mtlTaskContextList.add(new MtlTaskContext(MergeManifests.class));

                //mtlTaskContextList.add(new MtlTaskContext(PostProcessManifestTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(MergeManifestAwbsConfigAction.class, null));

                //mtlTaskContextList.add(new MtlTaskContext(MergeResV4Dir.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(ProcessAndroidResources.class));

                mtlTaskContextList.add(new MtlTaskContext(ProcessResAwbsTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(JavacAwbsTask.ConfigAction.class, null));

                if (null != androidExtension.getDataBinding() && androidExtension.getDataBinding().isEnabled()) {
                    mtlTaskContextList.add(new MtlTaskContext(AwbDataBindingRenameTask.ConfigAction.class, null));
                }

                mtlTaskContextList.add(new MtlTaskContext(TransformTask.class));

                if (appVariantContext.getAtlasExtension().getTBuildConfig().isIncremental() && (
                    appVariantContext.getBuildType().getPatchConfig() == null || !appVariantContext.getBuildType()
                        .getPatchConfig().isCreateTPatch())) {
                    mtlTaskContextList.add(new MtlTaskContext(PrepareBaseApkTask.ConfigAction.class, null));
                    final TaskFactory tasks = new TaskContainerAdaptor(project.getTasks());
                    VariantScope variantScope = appVariantContext.getVariantData().getScope();

                    // create the stream generated from this task
                    variantScope.getTransformManager().addStream(OriginalStream.builder().addContentType(
                        QualifiedContent.DefaultContentType.RESOURCES).addScope(QualifiedContent.Scope.PROJECT)
                                                                     .setFolders(new Supplier<Collection<File>>() {
                                                                         @Override
                                                                         public Collection<File> get() {
                                                                             return ImmutableList.of(new File(
                                                                                 appVariantContext.apContext
                                                                                     .getBaseApk() + "_"));
                                                                         }
                                                                     })
                                                                     // .setFolder(variantScope
                                                                     // .getSourceFoldersJavaResDestinationDir())
                                                                     // .setDependency(processJavaResourcesTask
                                                                     // .getName())
                                                                     .build());
                }

                mtlTaskContextList.add(new MtlTaskContext(PackageAwbsTask.ConfigAction.class, null));
                final TaskFactory tasks = new TaskContainerAdaptor(project.getTasks());
                VariantScope variantScope = appVariantContext.getVariantData().getScope();

                // create the stream generated from this task
                Builder builder = OriginalStream.builder();
                if (!appVariantContext.getAtlasExtension().getTBuildConfig().isIncremental() || (
                    appVariantContext.getBuildType().getPatchConfig() == null || !appVariantContext.getBuildType()
                        .getPatchConfig().isCreateTPatch())) {
                    builder.addContentType(ExtendedContentType.NATIVE_LIBS);
                }
                ;
                variantScope.getTransformManager().addStream(builder.addContentType(AtlasExtendedContentType.AWB_APKS)
                                                                 .addScope(QualifiedContent.Scope.PROJECT)
                                                                 .setFolders(new Supplier<Collection<File>>() {
                                                                     @Override
                                                                     public Collection<File> get() {
                                                                         return ImmutableList.of(
                                                                             appVariantContext.getAwbApkOutputDir());
                                                                     }
                                                                 })
                                                                 // .setFolder(variantScope
                                                                 // .getSourceFoldersJavaResDestinationDir())
                                                                 // .setDependency(processJavaResourcesTask.getName())
                                                                 .build());

                mtlTaskContextList.add(new MtlTaskContext("package"));

                mtlTaskContextList.add(new MtlTaskContext("zipalign"));

                mtlTaskContextList.add(new MtlTaskContext(ApBuildTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(DiffBundleInfoTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(TPatchDiffResAPBuildTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(TPatchDiffApkBuildTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext(TPatchTask.ConfigAction.class, null));

                mtlTaskContextList.add(new MtlTaskContext("assemble"));

                new MtlTaskInjector(appVariantContext).injectTasks(mtlTaskContextList, tAndroidBuilder);

                List<MtlTransformContext> mtlTransformContextList = new ArrayList<MtlTransformContext>();

                if (atlasExtension.getTBuildConfig().getClassInject()) {
                    mtlTransformContextList.add(
                        new MtlTransformContext(ClassInjectTransform.class, ProGuardTransform.class,
                                                DexTransform.class));
                }

                if (!mtlTransformContextList.isEmpty()) {
                    new MtlTransformInjector(appVariantContext).injectTasks(mtlTransformContextList);
                }

                if (atlasExtension.getTBuildConfig().isAtlasMultiDex()) {
                    List<BaseVariantOutputData> baseVariantOutputDataList = appVariantContext.getVariantOutputData();
                    for (final BaseVariantOutputData vod : baseVariantOutputDataList) {
                        TransformManager.replaceTransformTask(appVariantContext, vod, MultiDexTransform.class,
                                                              AtlasMultiDexTransform.class);
                    }
                }

                List<BaseVariantOutputData> baseVariantOutputDataList = appVariantContext.getVariantOutputData();
                for (final BaseVariantOutputData vod : baseVariantOutputDataList) {
                    TransformManager.replaceTransformTask(appVariantContext, vod, ProGuardTransform.class,
                                                          AtlasProguardTransform.class);
                    if (atlasExtension.getTBuildConfig().isIncremental()) {
                        final VariantOutputScope variantOutputScope = vod.getScope();
                        InstantRunPatchingPolicy patchingPolicy = variantScope.getInstantRunBuildContext()
                            .getPatchingPolicy();
                        AppVariantOutputContext appVariantOutputContext = appVariantContext.getAppVariantOutputContext(
                            vod);
                        AtlasDependencyTree atlasDependencyTree = AtlasBuildContext.androidDependencyTrees.get(
                            appVariantOutputContext.getVariantContext().
                                getVariantConfiguration().getFullName());

                        for (AwbBundle awbBundle : atlasDependencyTree.getAwbBundles()) {
                            AndroidTask<Dex> dexTask = androidTasks.create(tasks, new Dex.ConfigAction(
                                variantOutputScope.getVariantScope(), appVariantOutputContext, awbBundle));
                            dexTask.dependsOn(tasks, Iterables
                                .getLast(TaskQueryHelper.findTask(project, TransformTask.class, vod.variantData)));
                            PackagingScope packagingScope = new AwbPackagingScope(variantOutputScope, appVariantContext,
                                                                                  awbBundle);
                            AndroidTask<PackageAwb> packageAwb = androidTasks.create(tasks,
                                                                                     new PackageAwb
                                                                                         .StandardConfigAction(
                                                                                         packagingScope,
                                                                                         patchingPolicy));

                            packageAwb.dependsOn(tasks, dexTask);
                            PackageAwbsTask packageAwbsTask = Iterators.getOnlyElement(
                                TaskQueryHelper.findTask(project, PackageAwbsTask.class, vod.variantData).iterator());
                            packageAwbsTask.setEnabled(false);
                            packageAwbsTask.dependsOn(packageAwb.get(tasks));
                        }
                        AndroidTask<AwoInstallTask> awoInstallTask = androidTasks.create(tasks,
                                                                                         new AwoInstallTask
                                                                                             .ConfigAction(
                                                                                             appVariantContext, vod));
                        awoInstallTask.dependsOn(tasks, variantOutputScope.getVariantScope().getPackageApplicationTask()
                            .getName());
                    }
                }

                try {
                    hookFastMultiDex(appVariantContext);
                } catch (Exception e) {
                    throw new GradleException(e.getMessage(), e);
                }

            }
        });
    }

    private void hookFastMultiDex(AppVariantContext appVariantContext) throws Exception {

        if (null == appVariantContext.getAtlasExtension().getMultiDexConfigs()){
            return;
        }

        FastMultiDexer fastMultiDexer = new FastMultiDexer(appVariantContext);
        if (fastMultiDexer.isFastMultiDexEnabled()) {
            List<TransformTask> list = TransformManager.findTransformTaskByTransformType(appVariantContext,
                                                                                         DexTransform.class);
            for (TransformTask transformTask : list) {

                DefaultDexOptions dexOptions = (DefaultDexOptions)ReflectUtils.getField(transformTask.getTransform(),
                                                                                        "dexOptions");
                dexOptions.setPreDexLibraries(false);

                VariantScope variantScope = appVariantContext.getScope();
                GradleVariantConfiguration config = appVariantContext.getVariantData().getVariantConfiguration();

                DexTransform newDexTransform = new DexTransform(dexOptions, config.getBuildType().isDebuggable(), false,
                                                                null, variantScope.getPreDexOutputDir(),
                                                                AtlasBuildContext.androidBuilderMap
                                                                    .get(transformTask.getProject()),
                                                                appVariantContext.getProject().getLogger(),
                                                                variantScope.getInstantRunBuildContext(),
                                                                AndroidGradleOptions.getBuildCache(
                                                                    variantScope.getGlobalScope().getProject()));

                TransformManager.replaceTransformTask(transformTask, newDexTransform);

                boolean preDexLibraries = dexOptions.getPreDexLibraries();

                transformTask.doFirst(new Action<Task>() {
                    @Override
                    public void execute(Task task) {
                        AtlasBuildContext.androidBuilderMap.get(task.getProject()).multiDexer = fastMultiDexer;
                    }
                });

                transformTask.doLast(new Action<Task>() {
                    @Override
                    public void execute(Task task) {
                        AtlasBuildContext.androidBuilderMap.get(task.getProject()).multiDexer = null;
                    }
                });
            }
        }
    }
}
